mod error;

use symphonia::core::audio::{RawSampleBuffer, SignalSpec};
use symphonia::core::codecs::{Decoder, DecoderOptions, CODEC_TYPE_NULL};
use symphonia::core::errors::Error as SymError;
use symphonia::core::formats::{FormatReader, FormatOptions};
use symphonia::core::io::MediaSourceStream;
use symphonia::core::meta::MetadataOptions;
use symphonia::core::probe::Hint;
use tool_helper::{Error, Input};

type XASamples = RawSampleBuffer::<i16>;

pub fn load_for_xa_audio(input: Input) -> Result<XASamples, Error> {
    let media_stream = MediaSourceStream::new(Box::new(load_to_ram(input)?), Default::default());
    let format       = symphonia::default::get_probe().format(&Hint::new(), media_stream, &FormatOptions::default(), &MetadataOptions::default()).map_err(error::probe)?.format;
    let track        = format.tracks().iter().find(|t| t.codec_params.codec != CODEC_TYPE_NULL).ok_or_else(error::find_track)?;
    
    // Create a decoder for the track.
    let decoder  = symphonia::default::get_codecs().make(&track.codec_params, &DecoderOptions::default()).map_err(error::decoder)?;
    let track_id = track.id;

    decode(format, decoder, track_id)
}

fn decode(mut format: Box<dyn FormatReader>, mut decoder: Box<dyn Decoder>, track_id: u32) -> Result<XASamples, Error> {
    let mut xa_audio = None;

    loop {
        // Get the next packet from the media format.
        let packet = match format.next_packet() {
            Ok(packet) => packet,
            Err(err) => {
                if let SymError::IoError(io_err) = &err {
                    if io_err.kind() == std::io::ErrorKind::UnexpectedEof {
                        return xa_audio.ok_or_else(error::no_buffer_created);
                    }                    
                }
                return Err(error::next_packet(err));
            }
        };

        // Consume any new metadata that has been read since the last packet.
        format.metadata().skip_to_latest();

        // If the packet does not belong to the selected track, skip over it.
        if packet.track_id() != track_id {
            continue;
        }

        // Decode the packet into audio samples.
        let packet = decoder.decode(&packet).map_err(error::decode)?;
        if xa_audio.is_none() {
            let duration = packet.capacity() as u64;
            
            xa_audio = Some(XASamples::new(duration, SignalSpec::new_with_layout(44_100, symphonia::core::audio::Layout::Stereo)));
        }

        if let Some(xa_audio) = &mut xa_audio {
            xa_audio.copy_interleaved_ref(packet);
        }        
    }
}

fn load_to_ram(mut input: Input) -> Result<std::io::Cursor<Vec<u8>>, Error> {
    let mut buffer = Vec::default();
    
    input.read_to_end(&mut buffer)?;
    Ok(std::io::Cursor::new(buffer))
}